/*
 *  authbind.h - various definitions
 *
 *  authbind is Copyright (C) 1998 Ian Jackson
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 * 
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
 * 
 */

#ifndef AUTHBIND_H
#define AUTHBIND_H

#ifndef LIBAUTHBIND
# define LIBAUTHBIND "/usr/local/lib/authbind/libauthbind.so." MAJOR_VER
#endif

#ifndef HELPER
# define HELPER "/usr/local/lib/authbind/helper"
#endif

#define PRELOAD_VAR "LD_PRELOAD"
#define AUTHBINDLIB_VAR "AUTHBIND_LIB"
#define AUTHBIND_LEVELS_VAR "AUTHBIND_LEVELS"

#endif
